from .modules import Module


class ReLU(Module):
    """Ol' ReLU-iable."""

    def forward(self, xs):
        return xs.applymap(lambda x: max(x, 0))


class LayerMaxNorm(Module):
    """Normalize across the feature dimension with respect to the infinity norm."""

    def forward(self, xs):
        return xs.divide(xs.abs().max(axis=0), axis=1).fillna(0.0)



from abc import abstractmethod

import pandas as pd

from . import math


class Module(object):
    """An object that is callable via its .forward method."""

    def __call__(self, xs):
        return self.forward(xs)

    @abstractmethod
    def forward(self, xs):
        """Minimally, define this method to define a Module."""
        raise NotImplementedError

    def show(self):
        return str(self.__class__.__name__)


class Sequential(Module):
    """A Module that applies an iterable of Modules sequentially."""

    def __init__(self, *modules):
        self.modules = modules

    def forward(self, xs):
        for module in self:
            xs = module(xs)
        return xs

    def show(self):
        return f"{super().show()}{[self.show(module) for module in self]}"

    def __iter__(self):
        return iter(self.modules)

    def __getitem__(self, key):
        return self.modules[key]


class Linear(Module):
    """A Module that multiplies its inputs by the weights_df and adds the bias_series.

    Input 'tensors'  can be at most 2-D here: feature (rows) and batch/sequence (columns).

    The weights dataframe should have the input feature space as its column index
    and the output feature space as its row index."""

    def __init__(self, weights_df, bias_series=-1):
        self.w = weights_df
        if not isinstance(bias_series, pd.Series):
            bias_series = pd.Series(
                [bias_series] * len(weights_df), index=weights_df.index
            )
        self.b = bias_series

    def forward(self, xs):
        return (self.w @ xs).add(self.b, axis=0)

    def show(self):
        return f"{super().show()}:: {self.w.columns.name} -> {self.w.index.name}"


class AdditiveSkip(Module):
    """A Module that applies an additive "skip" connection around the provided Module."""

    def __init__(self, block):
        self.block = block

    def forward(self, xs):
        return xs + self.block(xs)

    def show(self):
        return f"[{super().show()}[{self.block.show()}]"


class Attention(Module):
    """Literally all you need, so we're done here."""

    def __init__(self, queries_df, keys_df, values_df):
        self.w_q = queries_df
        self.w_k = keys_df
        self.w_v = values_df

    def forward(self, xs):
        Q = self.w_q @ xs
        K = self.w_k @ xs
        V = self.w_v @ xs

        return Q, K, V


class LinearAttention(Attention):
    """The most basic version of an attention layer."""

    def forward(self, xs):
        V, Q, K = super().forward(xs)

        A = Q.T @ K

        zs = V @ A
        return zs


class SoftmaxAttention(Attention):
    """The best-known version of an attention layer."""

    def forward(self, xs):
        V, Q, K = super().forward(xs)

        A = math.softmax(Q.T @ K)

        zs = V @ A
        return zs



import numpy as np
import pandas as pd


def softmax(*args, **kwargs):
    raise NotImplementedError


def entropy(*args, **kwargs):
    raise NotImplementedError


def squared_difference(dfX, dfY):
    return (dfX - dfY).applymap(lambda r: r**2)


def eye(index):
    return pd.DataFrame(np.eye(len(index)), index=index, columns=index)



from .functional import LayerMaxNorm, ReLU
from .modules import AdditiveSkip, Linear, LinearAttention, Module
from .modules import Sequential, SoftmaxAttention

__all__ = [
    "Module",
    "Sequential",
    "AdditiveSkip",
    "Linear",
    "LinearAttention",
    "SoftmaxAttention",
    "LayerMaxNorm",
    "ReLU",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['panndas', 'panndas.data', 'panndas.nn']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3,<2.0']

setup_kwargs = {
    'name': 'panndas',
    'version': '0.0.0',
    'description': 'Neural networks in pandas',
    'long_description': '<div align="center">\n  <!-- <img src="https://pandas.pydata.org/static/img/pandas.svg"><br> -->\n</div>\n\n-----------------\n\n# panndas: neural networks in pandas\n[![PyPI Latest Release](https://img.shields.io/pypi/v/panndas.svg)](https://pypi.org/project/panndas/)\n[![Package Status](https://img.shields.io/pypi/status/panndas.svg)](https://pypi.org/project/panndas/)\n[![License](https://img.shields.io/pypi/l/panndas.svg)](https://github.com/pandas-dev/pandas/blob/main/LICENSE)\n[![Coverage](https://codecov.io/github/charlesfrye/panndas/coverage.svg?branch=main)](https://codecov.io/gh/charlesfrye/panndas)\n[![Downloads](https://static.pepy.tech/personalized-badge/panndas?period=month&units=international_system&left_color=black&right_color=orange&left_text=PyPI%20downloads%20per%20month)](https://pepy.tech/project/panndas)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## What is it?\n\n**pandas** is a Python package that provides fast, flexible, and expressive data\nstructures designed to make working with "relational" or "labeled" data both\neasy and intuitive.\n\n**panndas** is a neural network library built on top of pandas as a joke.\n\n## Main Features\nHere are just a few of the things that panndas does well:\n\n  -\n\n## License\n[YOLO](LICENSE)\n\n## Contributing to panndas [![Open Source Helpers](https://www.codetriage.com/pandas-dev/pandas/badges/users.svg)](https://www.codetriage.com/pandas-dev/pandas)\n\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are unwelcome.\n',
    'author': 'Charles Frye',
    'author_email': 'cfrye59@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/charlesfrye/panndas',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0.0',
}


setup(**setup_kwargs)
